/*
=================
Interactive Environment DX10 Win32 implementation
- Based on the DirectX SDK's "ShadowVolume10" sample [Copyright (c) Microsoft Corporation. All rights reserved.] 
- Please see the accompanying readme.txt
=================
*/
#include "DXUT.h"
#include "DXUTgui.h"
#include "DXUTsettingsdlg.h"
#include "DXUTcamera.h"
#include "SDKmisc.h"
#include "SDKmesh.h"
#include "resource.h"
#include "GenShadowMesh.h"

#define IDC_ENABLEFULLSCREENMODE 1

/*
=======
Globals
=======
*/

/* structure to hold vertex coordinates */
struct QuadVertex
{
    D3DXVECTOR3 VertexPosition;
};

/* structure used to initialise the coordinates and colour of each light source */
struct LightingProperties
{
    D3DXVECTOR3 Position;
    D3DXVECTOR4 Colour;

    LightingProperties() { }
    
	LightingProperties(D3DXVECTOR3 Position_, D3DXVECTOR4 Colour_)
    {
        Position = Position_;
        Colour = Colour_;
    }
};

/* setup the spatial position and colour of the light source */
LightingProperties g_SetupLights = LightingProperties(
										D3DXVECTOR3(-4.0f, 2.0f, -4.0f), 
										D3DXVECTOR4(10.0f, 10.0f, 10.0f, 1.0f));

/* structure storing the data of each light source */
struct LightData
{
    D3DXVECTOR3 m_LightPosition;
    D3DXVECTOR4 m_LightColour;
    D3DXMATRIX m_WorldTransformationMatrix;
};

/*
=================
Global Variables
=================
*/

/* declare the LightData object */
LightData g_LightObjectData;

//first-person perspective model view camera
CFirstPersonCamera g_FPSModelViewCamera;

//camera for controlling the 3-D mesh movement
CModelViewerCamera g_MeshControlCamera;

//camera for controlling light movement
CModelViewerCamera g_LightControlCamera;

//matrix to scale the object mesh (drone)
D3DXMATRIX g_MeshScalingMatrix;

//matrix to scale the map mesh (room)
D3DXMATRIX g_BackgroundWorldMeshMatrix;

//ambient lighting colour of the scene
D3DXVECTOR4 AmbientLighting(0.1f, 0.1f, 0.1f, 1.0f); 

//true when the left mouse button is pressed
bool g_bLeftMBPressed = false;

//true when the right mouse button is pressed
bool g_bRightMBPressed = false;

//true when the middle mouse button is pressed
bool g_bMiddleMBPressed = false;

/////////////////////////
//Direct3D 10 resources//
/////////////////////////

/* a mesh object used to read the background .sdkmesh files into memory */
CDXUTSDKMesh g_GameLevelMesh10[2];

/* a mesh object used to read the movable .sdkmesh file into memory */
CDXUTSDKMesh g_MeshObject;

//interface for implementing a rendering effect
ID3D10Effect* g_pID3D10Effect = NULL;

//interface managing a vertex buffer resource
ID3D10Buffer* g_pID3D10VertexBuffer = NULL; 

//interface for a vertex input layout object
ID3D10InputLayout* g_pID3D10VertexLayout = NULL; 

//ID3D10EffectTechnique interfaces
ID3D10EffectTechnique* g_pID3D10EffectRenderTextured = NULL;
ID3D10EffectTechnique* g_pID3D10EffectRenderLit = NULL;
ID3D10EffectTechnique* g_pID3D10EffectRenderAmbient = NULL;
ID3D10EffectTechnique* g_pID3D10EffectRenderShadow = NULL;

//DXUT helper for the management of shared dialog resources
CDXUTDialogResourceManager g_DXUTDialogResourceManager;

//the rendering device settings dialog
CD3DSettingsDlg g_D3DDeviceSettingsDialog;

//add and manage HUD elements
CDXUTDialog	g_3DUserInterfaceHUD;

//add and control specific HUD elements
CDXUTDialog	g_GUIControl;

/* ID3D10EffectMatrixVariable interfaces for reading shader
   variables as matrix types */

//projection matrix
ID3D10EffectMatrixVariable* g_pd3d10ProjMatrixVar = NULL; 

//view matrix
ID3D10EffectMatrixVariable* g_pd3d10ViewMatrixVar = NULL;

//world matrix
ID3D10EffectMatrixVariable* g_pd3d10WorldMatrixVar = NULL;

/* ID3D10EffectShaderResourceVariable interface for accessing
   shader-resource variables */
ID3D10EffectShaderResourceVariable* g_pd3d10DiffuseTexture = NULL;

/* ID3D10EffectVectorVariable interfaces for accessing shader
   variables as vector types */
ID3D10EffectVectorVariable* g_pd3d10LightPositionVectorVar = NULL;
ID3D10EffectVectorVariable* g_pd3d10LightColourVectorVar = NULL;
ID3D10EffectVectorVariable* g_pd3d10AmbientLightingVectorVar = NULL;
ID3D10EffectVectorVariable* g_pd3d10ShadowColourVectorVar = NULL;

/* ID3D10EffectScalarVariable interfaces for accessing scalar
   shader-resource variables */
ID3D10EffectScalarVariable* g_pd3d10ExtrudeShadowAmountScalarVar = NULL;
ID3D10EffectScalarVariable* g_pd3d10ExtrudeShadowBiasScalarVar = NULL;

int g_BackgroundSceneMesh;

/*
=================
Function Prototypes 
=================
*/
void Initialise();

LRESULT CALLBACK MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbGUIState, void* pUserContext);

HRESULT CALLBACK OnD3D10CreateDevice(ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext);
HRESULT CALLBACK OnD3D10SwapChainResized(ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext);

void CALLBACK MouseEventProcessing(bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos, void* pUserContext);
void CALLBACK GUIEventCallbackProcessing(UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext);

bool CALLBACK SetD3D10DeviceSettings(DXUTDeviceSettings* pDeviceSettings, void* pUserContext);
void CALLBACK HandleSceneFrameUpdates(double fTime, float fElapsedTime, void* pUserContext);
void CALLBACK RenderFrame(ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext);
void CALLBACK ReleaseSwapChain(void* pUserContext);

void CALLBACK OnD3D10DestroyDevice(void* pUserContext);

/*
================
wWinMain
-entry point to the application
-initialise the message processing loop elements and the idle time required for the rendering of our scene 
================
*/
int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow)
{
    /* set DXUT callbacks */
	////////////////////////

	//sets a callback function to change the device settings prior to device creation
    DXUTSetCallbackDeviceChanging(SetD3D10DeviceSettings );
	//sets the main message callback function
    DXUTSetCallbackMsgProc(MsgProc);

    //sets the mouse event callback function
	DXUTSetCallbackMouse(MouseEventProcessing);
    //sets the frame update callback function
	DXUTSetCallbackFrameMove(HandleSceneFrameUpdates);

	//sets callback creating Direct3D 10 resources not dependent on the back buffer
    DXUTSetCallbackD3D10DeviceCreated(OnD3D10CreateDevice);
	//sets callback creating Direct3D 10 resources dependent on the back buffer
    DXUTSetCallbackD3D10SwapChainResized(OnD3D10SwapChainResized);
    //sets the callback function to release resources created in the OnD3D10ResizedSwapChain function
	DXUTSetCallbackD3D10SwapChainReleasing(ReleaseSwapChain);
	//sets the callback function to release resources created in the nD3D10CreateDevice function
    DXUTSetCallbackD3D10DeviceDestroyed(OnD3D10DestroyDevice);
    
	//sets the callback function rendering the scene on a per-frame basis
	DXUTSetCallbackD3D10FrameRender(RenderFrame);

	//initialise the application
    Initialise();
    
	//initialises DXUT: parses for command line arguments, shows a message box on errors
	DXUTInit(true, true, NULL);
	//options for mouse cursor in full-screen mode (show it and prevent it from exiting the screen boundaries)
	DXUTSetCursorSettings(true, true);
	//creates an application window with the specified caption
	DXUTCreateWindow(L"An Interactive Environment");
    //create a Direct3D 10 device starting in windowed mode with the specified width and height
	DXUTCreateDevice(true, 1024, 768);
    //enter the main DXUT execution loop
	DXUTMainLoop();

    return DXUTGetExitCode();
}

/*
================
Initialise
-initialises the application HUD, cameras, lights and the scaling and translation of the map mesh
================
*/
void Initialise()
{
	/* init the application HUD (the"Full-Screen Mode" button) */
    g_D3DDeviceSettingsDialog.Init(&g_DXUTDialogResourceManager);
    g_3DUserInterfaceHUD.Init(&g_DXUTDialogResourceManager);
    g_GUIControl.Init(&g_DXUTDialogResourceManager);

    g_3DUserInterfaceHUD.SetCallback(GUIEventCallbackProcessing ); 
    g_3DUserInterfaceHUD.AddButton(IDC_ENABLEFULLSCREENMODE, L"Full-Screen Mode", 35, 10, 125, 22);
    
    g_GUIControl.SetCallback(GUIEventCallbackProcessing); 
	
    g_FPSModelViewCamera.SetRotateButtons( true, false, false );
    g_MeshControlCamera.SetButtonMasks( MOUSE_RIGHT_BUTTON, 0, 0 );
    g_LightControlCamera.SetButtonMasks( MOUSE_MIDDLE_BUTTON, 0, 0 );

    //init the light
    g_LightObjectData.m_LightPosition = g_SetupLights.Position;
    g_LightObjectData.m_LightColour = g_SetupLights.Colour;

    /* scale and translate the environment's map mesh */
	////////////////////////////////////////////////////

	//the translation matrix
    D3DXMATRIX mapTranslationMatrix;

	//create the translation matrix
    D3DXMatrixTranslation(&g_BackgroundWorldMeshMatrix, 0.0f, 1.0f, 0.0f);
    D3DXMatrixTranslation(&mapTranslationMatrix, 1.0f, 1.0f, 0.0f);

	//create an identity matrix
    D3DXMatrixIdentity(&g_MeshScalingMatrix);
}

/*
================
SetD3D10DeviceSettings
-specify how to create the D3D10 device
================
*/
bool CALLBACK SetD3D10DeviceSettings(DXUTDeviceSettings* pDeviceSettings, void* pUserContext)
{
	/* the DXGI_FORMAT_D24_UNORM_S8_UINT format supports
	   stencilling */
    pDeviceSettings->d3d10.AutoDepthStencilFormat = DXGI_FORMAT_D24_UNORM_S8_UINT;

    return true;
}

/*
================
HandleSceneFrameUpdates
-processes each scene update
================
*/
void CALLBACK HandleSceneFrameUpdates( double time, float timePassed, void* context)
{
	/* update the view matrix based on user input and elapsed time */
    g_FPSModelViewCamera.FrameMove(timePassed);
    g_MeshControlCamera.FrameMove(timePassed);
    g_LightControlCamera.FrameMove(timePassed);
}

/*
================
MsgProc
-handle all application messages (callback function is called whenever an event occurs)
================
*/
LRESULT CALLBACK MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbGUIState, void* pUserContext )
{
	/* first let the dialogs handle all generated messages 
	   before passing on the remaining messages */
	//////////////////////////////////////////////
    
	//update the GUI state
	*pbGUIState = g_DXUTDialogResourceManager.MsgProc(hWnd, uMsg, wParam, lParam);
    if(*pbGUIState)
        return 0;

    /* handle generated msgs as a priority */
    *pbGUIState = g_3DUserInterfaceHUD.MsgProc(hWnd, uMsg, wParam, lParam);
    if(*pbGUIState)
        return 0;
    
	*pbGUIState = g_GUIControl.MsgProc(hWnd, uMsg, wParam, lParam);
    if(*pbGUIState)
        return 0;

	/* all remaining messages (user input) should be passed to
	   the camera */
    g_FPSModelViewCamera.HandleMessages(hWnd, uMsg, wParam, lParam);
    g_MeshControlCamera.HandleMessages(hWnd, uMsg, wParam, lParam);
    g_LightControlCamera.HandleMessages(hWnd, uMsg, wParam, lParam);

    return 0;
}

/*
================
MouseEventProcessing
-processes mouse input
================
*/
void CALLBACK MouseEventProcessing(bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos, void* pUserContext)
{
	/* flags indicating which mouse buttons are pressed */
    bool bOldLeftButtonDown = g_bLeftMBPressed;
    bool bOldRightButtonDown = g_bRightMBPressed;
    bool bOldMiddleButtonDown = g_bMiddleMBPressed;
    
	//is the left mouse button down?
	g_bLeftMBPressed = bLeftButtonDown;
	//is the middle mouse button down?
    g_bMiddleMBPressed = bMiddleButtonDown;
	//is the right mouse button down?
    g_bRightMBPressed = bRightButtonDown;

	//move the mesh if the right mouse button is down
    if(bOldRightButtonDown && !g_bRightMBPressed)
    {
        g_MeshControlCamera.SetEnablePositionMovement(false);
    }
    else 
		if(!bOldRightButtonDown && g_bRightMBPressed)
		{
			g_MeshControlCamera.SetEnablePositionMovement(true);
			g_FPSModelViewCamera.SetEnablePositionMovement(false);
		}

	//rotate the player camera if the left mouse button is down
    if(bOldLeftButtonDown && !g_bLeftMBPressed)
        g_FPSModelViewCamera.SetEnablePositionMovement(false);
    else 
		if(!bOldLeftButtonDown && g_bLeftMBPressed)
			g_FPSModelViewCamera.SetEnablePositionMovement(true);

	//move the light source if the middle mouse button is down
    if(bOldMiddleButtonDown && !g_bMiddleMBPressed)
    {
        g_LightControlCamera.SetEnablePositionMovement(false);
    } 
    else 
		if(!bOldMiddleButtonDown && g_bMiddleMBPressed)
		{
			g_LightControlCamera.SetEnablePositionMovement(true);
			g_FPSModelViewCamera.SetEnablePositionMovement(false);
		}

    /* move the player camera if none of the mouse buttons are
	   held down */
    if(!g_bRightMBPressed && !g_bMiddleMBPressed && !g_bLeftMBPressed)
        g_FPSModelViewCamera.SetEnablePositionMovement(true);
}

/*
================
GUIEventCallbackProcessing
-processes GUI events (selection of the full-screen mode button)
////TODO: Fix so that only Full Screen Mode can be set...cover net dit in die chapter discussion
================
*/
void CALLBACK GUIEventCallbackProcessing(UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext)
{   
    switch(nControlID)
    {
        case IDC_ENABLEFULLSCREENMODE: DXUTToggleFullScreen(); 
			break;
    }
}

/*
================
OnD3D10CreateDevice
-create Direct3D 10 resources not dependent on the back buffer
================
*/
HRESULT CALLBACK OnD3D10CreateDevice(ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC *pBackBufferSurfaceDesc, void* pUserContext)
{
    g_DXUTDialogResourceManager.OnD3D10CreateDevice(pd3dDevice);
    g_D3DDeviceSettingsDialog.OnD3D10CreateDevice(pd3dDevice);
    
    //the effect file
    WCHAR effectName[MAX_PATH];

	//read and compile the effect
    DXUTFindDXSDKMediaFileCch(effectName, MAX_PATH, L"MainFX10.fx");
    
	//create an effect from the file
    D3DX10CreateEffectFromFile(effectName, NULL, NULL, "fx_4_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, pd3dDevice, NULL, NULL, &g_pID3D10Effect, NULL, NULL);

    /* get the technique handles by name from the MainFX10.fx file */
    g_pID3D10EffectRenderTextured = g_pID3D10Effect->GetTechniqueByName("RenderTextured");
    g_pID3D10EffectRenderLit = g_pID3D10Effect->GetTechniqueByName("RenderLitEnvironment");
    g_pID3D10EffectRenderAmbient = g_pID3D10Effect->GetTechniqueByName("RenderWithAmbientLighting");
    g_pID3D10EffectRenderShadow = g_pID3D10Effect->GetTechniqueByName("RenderSceneWithShadow");
    
	/* create the input-assembler stages single element description */
    const D3D10_INPUT_ELEMENT_DESC vertex_input_layout[] =
    {
        {"POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, 
		  D3D10_INPUT_PER_VERTEX_DATA, 0},
        {"TEXTURE", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 24, 
		  D3D10_INPUT_PER_VERTEX_DATA, 0},
		{"NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, 
		  D3D10_INPUT_PER_VERTEX_DATA, 0},
    };

	//structure to describe each effect pass
    D3D10_PASS_DESC EffectPassDescription;
    
	//get the effect pass to render the scene lit
	g_pID3D10EffectRenderLit->GetPassByIndex(0)->GetDesc(&EffectPassDescription);
    
	//create an input-layout object
	pd3dDevice->CreateInputLayout(vertex_input_layout, 3, EffectPassDescription.pIAInputSignature, EffectPassDescription.IAInputSignatureSize, &g_pID3D10VertexLayout);

    //load the mesh representing the environment/game map and the character mesh
    g_GameLevelMesh10[0].Create(pd3dDevice, L"\\blackholeroom.sdkmesh", false, true);
    g_MeshObject.Create(pd3dDevice, L"\\EvilDrone.sdkmesh", false, true);

	//get the effect variables by name (from the MainFX10.fx file)
    g_pd3d10ProjMatrixVar = g_pID3D10Effect->GetVariableByName("ProjectionMatrix")->AsMatrix();
    g_pd3d10ViewMatrixVar = g_pID3D10Effect->GetVariableByName("ViewMatrix")->AsMatrix();
    g_pd3d10WorldMatrixVar = g_pID3D10Effect->GetVariableByName("WorldMatrix")->AsMatrix();
    g_pd3d10DiffuseTexture = g_pID3D10Effect->GetVariableByName("DiffuseTexture")->AsShaderResource();
    g_pd3d10LightPositionVectorVar = g_pID3D10Effect->GetVariableByName("LightPosition")->AsVector();
    g_pd3d10LightColourVectorVar = g_pID3D10Effect->GetVariableByName("LightColour")->AsVector();
    g_pd3d10AmbientLightingVectorVar = g_pID3D10Effect->GetVariableByName("AmbientLighting")->AsVector();
    g_pd3d10ShadowColourVectorVar = g_pID3D10Effect->GetVariableByName("ShadowColour")->AsVector();
    g_pd3d10ExtrudeShadowAmountScalarVar  = g_pID3D10Effect->GetVariableByName("ShadowExtrusionAmount")->AsScalar();
    g_pd3d10ExtrudeShadowBiasScalarVar = g_pID3D10Effect->GetVariableByName("ShadowExtrusionBias")->AsScalar();

    //set the camera at the centre of projection (eye) pointed towards the "at" location
	D3DXVECTOR3 eye(0.0f, 3.0f, -8.0f);
    D3DXVECTOR3 at(0.0f, 3.1f, 0.0f);
    g_FPSModelViewCamera.SetViewParams(&eye, &at);
    g_LightControlCamera.SetViewParams(&eye, &at);
    g_MeshControlCamera.SetViewParams(&eye, &at);

    return S_OK;
}

/*
================
OnD3D10SwapChainResized
-called whenever the swap chain resized event occurs
================
*/
HRESULT CALLBACK OnD3D10SwapChainResized(ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext)
{
	//calculate aspect ratio
	float WidthHeightRatio = pBackBufferSurfaceDesc->Width/(FLOAT)pBackBufferSurfaceDesc->Height;
	
	//called when the Direct3D 10 swap chain is about to be resized or created
    g_DXUTDialogResourceManager.OnD3D10ResizedSwapChain(pd3dDevice, pBackBufferSurfaceDesc);

    //set the camera's projection parameters
	g_FPSModelViewCamera.SetProjParams(D3DX_PI/4, WidthHeightRatio, 0.1f, 500.0f );
    g_MeshControlCamera.SetWindow(pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);
    g_LightControlCamera.SetWindow(pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);

    return S_OK;
}

/*
================
RenderScene
-setup the matrices and render the map and object mesh
================
*/
void RenderScene(ID3D10Device* pd3dDevice, ID3D10EffectTechnique* pEffectTechnique, bool renderShadowVol)
{
    // Setup the view matrices
    D3DXMATRIX ProjectionMatrix;
    D3DXMATRIX ViewMatrix;
    D3DXMATRIX ViewProjectionMatrix;
	D3DXMATRIX WorldMatrix;
    D3DXMATRIX WorldViewProjectionMatrix;
	
	//calculate the projection matrix
    ProjectionMatrix = *g_FPSModelViewCamera.GetProjMatrix();
    //calculate the view matrix
	ViewMatrix = *g_FPSModelViewCamera.GetViewMatrix();
	
	//calculate and set the view project matrix
    ViewProjectionMatrix = ViewMatrix * ProjectionMatrix;
    g_pd3d10ViewMatrixVar->SetMatrix( (float*)&ViewProjectionMatrix);

    /* render the mesh representing the map/level */
    if(!renderShadowVol)
    {
		//calculate and set the world view projection matrix
        WorldViewProjectionMatrix = g_BackgroundWorldMeshMatrix * ViewMatrix * ProjectionMatrix;
        g_pd3d10ProjMatrixVar->SetMatrix((float*)&WorldViewProjectionMatrix);
        g_pd3d10WorldMatrixVar->SetMatrix((float*)&g_BackgroundWorldMeshMatrix);
        
		//render the map mesh
		g_GameLevelMesh10[ g_BackgroundSceneMesh ].Render( pd3dDevice, pEffectTechnique, g_pd3d10DiffuseTexture );
    }

    /* render the mesh representing the object/character */
	///////////////////////////////////////////////////////
    
	//calculate the world matrix
	WorldMatrix = g_MeshScalingMatrix * (*g_MeshControlCamera.GetWorldMatrix());
    //calculate the world view projection matrix
	WorldViewProjectionMatrix = WorldMatrix * ViewMatrix * ProjectionMatrix;
    //set the world and world view project matrices
	g_pd3d10ProjMatrixVar->SetMatrix( (float*)&WorldViewProjectionMatrix );
    g_pd3d10WorldMatrixVar->SetMatrix( (float*)&WorldMatrix );
 
	//render the character mesh and the shadow
	if(renderShadowVol)
        g_MeshObject.RenderAdjacent(pd3dDevice, pEffectTechnique, g_pd3d10DiffuseTexture);
    else
        g_MeshObject.Render(pd3dDevice, pEffectTechnique, g_pd3d10DiffuseTexture);
}

/*
================
RenderFrame
-render a complete frame with meshes, shadows, lights, etc
================
*/
void CALLBACK RenderFrame(ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext)
{
	//set the clear colour to black
	float RenderTargetClearColour[4] = {0.0, 0.0, 0.0, 0.0};
    
	//clear the render target
	ID3D10RenderTargetView* pRenderTargetView = DXUTGetD3D10RenderTargetView();
    pd3dDevice->ClearRenderTargetView(pRenderTargetView, RenderTargetClearColour);
    
	//clear the stencil buffer
	ID3D10DepthStencilView* pDepthStencilView = DXUTGetD3D10DepthStencilView();
    pd3dDevice->ClearDepthStencilView(pDepthStencilView, D3D10_CLEAR_DEPTH, 1.0f, 0);

    //bind the input-layout object to the input-assembler stage
    pd3dDevice->IASetInputLayout(g_pID3D10VertexLayout);

    //draw the scene with ambient lighting
    g_pd3d10AmbientLightingVectorVar->SetFloatVector((float*)&AmbientLighting);
    RenderScene(pd3dDevice, g_pID3D10EffectRenderAmbient, false);
    
	//set the amount and bias to extrude the shadow volume from the silhouette edge
	g_pd3d10ExtrudeShadowAmountScalarVar ->SetFloat(120.0f - 0.1f);
    g_pd3d10ExtrudeShadowBiasScalarVar->SetFloat(0.1f);

	/* setup the light */
	D3DXVECTOR4 LightVector(g_LightObjectData.m_LightPosition.x, g_LightObjectData.m_LightPosition.y, g_LightObjectData.m_LightPosition.z, 1.0f);
    D3DXVec4Transform(&LightVector, &LightVector, g_LightControlCamera.GetWorldMatrix());
    g_pd3d10LightPositionVectorVar->SetFloatVector((float*)&LightVector);
    g_pd3d10LightColourVectorVar->SetFloatVector((float*)g_LightObjectData.m_LightColour);

	/*for the lightsource, render the resulting shadow*/
	/////////////////////////////////////////////////////
    
	//clear the stencil buffer
    pd3dDevice->ClearDepthStencilView(pDepthStencilView, D3D10_CLEAR_STENCIL, 1.0, 0);

    //prepare to render the shadow volume
    ID3D10EffectTechnique* pEffectTechnique = g_pID3D10EffectRenderShadow;
    pEffectTechnique = g_pID3D10EffectRenderShadow;

	//render the actual shadow
	RenderScene(pd3dDevice, pEffectTechnique, true);
    
	//render the scene with normal lighting
    RenderScene(pd3dDevice, g_pID3D10EffectRenderLit, false);

	/* code to render the GUI  the "Full-Screen Mode" button */
    g_3DUserInterfaceHUD.OnRender(fElapsedTime);
    g_GUIControl.OnRender(fElapsedTime);
    
	DXUT_EndPerfEvent();
}

/*
================
ReleaseSwapChain
-release resources created in OnD3D10ResizedSwapChain
================
*/
void CALLBACK ReleaseSwapChain( void* pUserContext )
{
    g_DXUTDialogResourceManager.OnD3D10ReleasingSwapChain();
}

/*
================
OnD3D10DestroyDevice
-release resources created in OnD3D10CreateDevice
================
*/
void CALLBACK OnD3D10DestroyDevice( void* pUserContext )
{
    g_DXUTDialogResourceManager.OnD3D10DestroyDevice();
    g_D3DDeviceSettingsDialog.OnD3D10DestroyDevice();
    DXUTGetGlobalResourceCache().OnDestroyDevice();
   
    SAFE_RELEASE( g_pID3D10Effect );
    SAFE_RELEASE( g_pID3D10VertexLayout );
    SAFE_RELEASE( g_pID3D10VertexBuffer );

    g_GameLevelMesh10[0].Destroy();
    g_GameLevelMesh10[1].Destroy();
    g_MeshObject.Destroy();
}

//eof