//OpenGL Win32 interactive environment

//#include <GL/glew.h> //see Linux/Unix version for glew usage
#include <GL/glut.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sstream> //included for stringstream
#include <string>

using namespace std;

#include "Geometry.h"
#include "VolumetricFog.h"
#include "ClothSimulation.h"

/* ASCII codes for various special Keys */
#define ESCAPE 27
#define UP_ARROW 72
#define DOWN_ARROW 80
#define LEFT_ARROW 75
#define RIGHT_ARROW 77

/* Macro Defines */
#define LIGHTDIR_H 100.0
#define LIGHTDIR_V 60.0
#define PI 3.1415926532L

/* Globals 
   - config setting, assigned values are defaults */
double g_curX = 21, g_curY = -1.5, g_curZ = 29;
double g_angle = 120;
double g_lightAngle = 0.0;
float g_frames = 0;
float g_timeZero = 0;
float temp = 0;

/* new global cloth instance */
Cloth *clothInstance = new Cloth();

/*
=================
SetupLights
-specify the lights for our scene
=================
*/
void SetupLights()
{
	/* ambient RGBA reflectance of the material */
	float materialAmbience[] = {0.5, 0.5, 0.5, 1.0};
	/* ambient RGBA intensity of entire scene */
	float sceneAmbience[] = {0.7, 0.7, 0.7, 1.0};

	/* specify lighting model parameters 
	   - GL_LIGHT_MODEL_AMBIENT: specify ambient RGBA intensity of entire scene*/
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, sceneAmbience);
	/* specify material parameters for lighting model 
	   - GL_FRONT_AND_BACK: front and back faces are being updated 
	   - GL_AMBIENT: update all the ambient faces */
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, materialAmbience);
}

/*
=================
SetModeTextured
-gets called to enable texturing
=================
*/
void SetModeTextured()
{
	/* enable texturing */
	glEnable(GL_TEXTURE_2D);
	/* if texturing isn't enabled, generate error */
	if(!glIsEnabled(GL_TEXTURE_2D))
		fprintf(stderr, "Texture enabling failed.\n");
}

/*
=================
setLightOff
-if called disables lighting
=================
*/
void SetLightsOff()
{
	glDisable(GL_LIGHTING);
}

/*
=================
setLightOn
-enables lighting
=================
*/
void SetLightsOn()
{
	glEnable(GL_LIGHTING);
}

/*
=================
SetSmoothShading
-enables smooth shading
=================
*/
void SetSmoothShading()
{

	glEnable(GL_LIGHTING);
	glShadeModel(GL_SMOOTH);
}

/*
=================
SetFlatShading
-enables flat shading
=================
*/
void SetFlatShading()
{
	glEnable(GL_LIGHTING);
	glShadeModel(GL_FLAT);
}

/*
=================
ActivateLight
-activates the GL_LIGHT0 light source
=================
*/
void ActivateLight()
{
	glEnable(GL_LIGHT0);
}

/*
=================
SetupLightDirection
-sets the light source position
=================
*/
void SetupLightDirection()
{
	float lightDirection[] = {sin(g_lightAngle) * LIGHTDIR_H, LIGHTDIR_V, cos(g_lightAngle) * LIGHTDIR_H, 0.0};
	glLightfv(GL_LIGHT0, GL_POSITION, lightDirection);
}

/*
=================
SwitchToOrtho
-switches from perspective to orthographic projection
NOTE: (0,0) is situated at the top left corner of your screen
=================
*/
void SwitchToOrtho()
{
	/* specifies the Projection matrix as the current matrix */
	glMatrixMode(GL_PROJECTION);
	/* pushes current matrix stack down by one */
	glPushMatrix();
	/* move to the center of the screen */
	glLoadIdentity();
	/* define a 2-D orthographic projection matrix */
	gluOrtho2D(0, 1024, 0, 768);
	/* scales along the x, -y, and z axes */
	glScalef(1, -1, 1);
	/* moves along the x, y and z-axis, moves -'768' units (down) on y-
       axis */
	glTranslatef(0, -1.0*768, 0);
	/* specifies the Modelview matrix as the current matrix */
	glMatrixMode(GL_MODELVIEW);
	/* move to the center of the screen */
	glLoadIdentity();
}

/*
=================
SwitchFromOrtho
-switches back from orthographic projection to perspective projection
=================
*/
void SwitchFromOrtho()
{
	/* specifies the Projection matrix as the current matrix */
	glMatrixMode(GL_PROJECTION);
	/* pops current matrix stack */
	glPopMatrix();
	/* specifies the Modelview matrix as the current matrix */
	glMatrixMode(GL_MODELVIEW);
}

/*
=================
Display
-calls all the geometric constructs as defined in geometry.h
=================
*/
void Display(void)
{
	/* clear frame buffer and z-buffer - thus clearing the screen */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    /* move to the center of the screen */
    glLoadIdentity();

    /* swith to orthographic projection, we are currently in perspective projection,
	   the text has to be rendered in orthographic projection mode */
	SwitchToOrtho();

	/* set text colour to white */
	glColor3f(1.0,1.0,1.0);
	/* sets the font position at (0, 20) */
	glRasterPos2i(0, 20);

	/* draws the FPS to the screen */
	/* --------------------------- */
	stringstream ss;
	/* temp calculated in the "Idle" function */
	ss << temp;  
	string s = ss.str();
	
	/* renders the FPS raster characters */
	for(int i = 0; s[i]; i++)
		glutBitmapCharacter(GLUT_BITMAP_8_BY_13, s[i]);
	/* --------------------------- */

	/* switch back to perspective projection */
	SwitchFromOrtho();

	/* pushes current matrix stack down by one */
	glPushMatrix();
		/* specifies the Modelview matrix as the current matrix */
		glMatrixMode(GL_MODELVIEW);

		/* rotate camera 'g_angle' degrees about the y-axis */
		glRotatef(g_angle, 0.0, 1.0, 0.0);

		/* translate (move) camera to the intitial "spawn" position, thereafter
		   moving the camera to the updated position as calculated by
		   InputKeyPressed */
		glTranslatef(g_curX, g_curY, g_curZ);

		/* inits the light source position */
		SetupLightDirection();

		/* calls to the world drawing routines as defined in geometry.h */
		/* ------------------------------------------------------------ */
		drawMainBottomFloor();
		constructPillarArchitecture();
		constructOuterWalls();
		constructInnerPedestal();
		constructTerraces();
		constructTerraceRailing();
		constructPersianRoof();
		constructRoofWalls();
		constructRoofWallWindow();
		constructPool();
		constructPoolSteps();
		constructSupportBeam();
		drawPoolBottomFog();
		drawPoolWater();
		/* ------------------------------------------------------------ */

		/* disable culling 
		   - without disabled culling cloth only renders facing triangles */
		glDisable(GL_CULL_FACE);
		/* initialise the cloth */
		clothInstance->drawCloth();
		/* enable culling (remove all the faces pointing away from the viewer) */
		glEnable(GL_CULL_FACE);
	/* pops current matrix stack */
	glPopMatrix();

	/* contents of back buffer are set to become that of front buffer (double
       buffering)*/
	glutSwapBuffers();
	/* make sure all previously issued commands are completed */
	glFlush();
}

/*
=================
Reshape
-rendered window's Reshape function - manages changes to size of the window, correctly adjusting aspect ratio.
=================
*/
void Reshape(int w, int h)
{
	/* sets the viewport with the first two parameters the lower-left corner of the viewport
	   rectangle, in pixels and the last two, the width and height of the viewport */
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);

	/* specifies the Projection matrix as the current matrix */
	glMatrixMode(GL_PROJECTION);
	/* move to the center of the screen */
	glLoadIdentity();

	/* sets up the  perspective projection matrix:
	   - the first parameter specifies the field of view (fov), an g_angle in degrees
	     (in the y-direction)
	   - the second parameter is aspect ratio determining fov
		 (in the x-direction)
	   - the third parameter is the distance from the viewer to the near clipping plane
	   - the last parameter is the distance from the viewer to the far clipping plane */
	gluPerspective(45, (GLdouble)w / (GLdouble)h, 0.05, 100);
	/* specifies the Modelview matrix as the current matrix */
	glMatrixMode(GL_MODELVIEW);
}

/*
=================
MyInit
-handles basic initialisation and setup of the OpenGL parameters
=================
*/
void MyInit(void)
{
	/* start by clearing the color buffers with black */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	
	/* our default is smooth shading */
	SetSmoothShading();

	/* enable hidden-surface removal  */
	glEnable(GL_DEPTH_TEST);
	/* culling enabled */
	glEnable(GL_CULL_FACE);
	
	/* setup the texture loader */
	textureLoader();
	
	/* enable normal normalization */
	glEnable(GL_NORMALIZE);
	/* specifies front- and back-facing polygons 
	   -counterclockwise polygons are front-facing (GL_CCW) 
	   -clockwise polygons are front-facing (GL_CW) */
	glFrontFace(GL_CCW);
	/* specifies whether front- or back-faces are to be culled */
    glCullFace(GL_BACK);
		
	/* initialises the volumetric fog - defined in VolumetricFog.h */
	initialiseVolumetricFogExtension();

	/* set the Fog parameters */
	/* fog fade is linear */
	glFogi(GL_FOG_MODE, GL_LINEAR);
	/* set the fog colour */
	glFogfv(GL_FOG_COLOR, fog_colour);
	/* set the fog start (Least Dense) */
	glFogf(GL_FOG_START,  0.0f);
	/* set the fog end (Most Dense) */
	glFogf(GL_FOG_END,    1.0f);
	/* per-pixel fog calculation */
	glHint(GL_FOG_HINT, GL_NICEST);
	/* set fog based on vertice coords - hence volumetric fog*/
	glFogi(GL_FOG_COORDINATE_SOURCE_EXT, GL_FOG_COORDINATE_EXT);

	/* activate the defined light source */
	ActivateLight();
	/* setup the light source */
	SetupLights();
	/* enable lighting */
	SetLightsOn();
	/* enable texturing */
	SetModeTextured();
}

/*
=================
InputKeyPressed
-Controls "player" movement: forward, backward, left, right
=================
*/
void InputKeyPressed(int key, int x, int y)
{
	/* switch control structure */
    switch (key)
	{
		/* when the up-arrow on the keyboard is pressed */
		case GLUT_KEY_UP:
		{
			double rads = g_angle / 180.0L * PI;
			/* the smaller the number divided by the faster you glide forward */
			double dx = -sin(rads) / 1.5;
			/* the smaller the number divided by the faster you glide forward */
			double dz = cos(rads) / 1.5;

			g_curX += dx;
			g_curZ += dz;
		}
		break;

		/* when the back-arrow on the keyboard is pressed */
		case GLUT_KEY_DOWN:
		{
			double rads = g_angle / 180.0L * PI;
			/* the smaller the number divided by the faster you glide forward */
			double dx = -sin(rads) / 3.0;
			/* the smaller the number divided by the faster you glide forward */
			double dz = cos(rads) / 3.0;

			g_curX -= dx;
			g_curZ -= dz;
		}
		break;

		/* when the right-arrow on the keyboard is pressed */
		case GLUT_KEY_RIGHT:
			g_angle += 5;
		break;

		/* when the left-arrow on the keyboard is pressed */
	   	case GLUT_KEY_LEFT:
			g_angle -= 5;
		break;

		/* when anything else is pressed */
   		default:
			break;
    }
	/* sets the current window for redisplay */
	glutPostRedisplay();
}

/*
=================
Keys
-Handles all normal key input
=================
*/
void Keys(unsigned char key, int x, int y)
{
	/* exit the program */
	if(key == 'q' || key == 'Q')
	{
		exit(-1);
	}

	/* move vertically up */
	if(key == 'w' || key == 'W')
		g_curY -= 2;

	/* move vertically down */
	if(key == 's' || key == 'S')
		g_curY += 2;
}

/*
=================
Idle
-used for animation of our scene
=================
*/
void Idle()
{
	/* set the initial wind force blowing on the cloths */
	clothInstance->wind = 0.4;
	clothInstance->clothPhysics();
	clothInstance->initUnitForce();

	/* calculate the frames per second rendered */
	/* ---------------------------------------- */
	g_frames++;
	float time = glutGet(GLUT_ELAPSED_TIME);
	if (time - g_timeZero > 1000)
	{
		temp = (g_frames*1000/(time-g_timeZero));

		glRasterPos2i(3, 20);

		printf("%f\n",temp); //debug console output
		g_timeZero = time;
		g_frames = 0;
	}
	/* ---------------------------------------- */

	/* sets the current window for redisplay */
	glutPostRedisplay();

	/* draw wireframe or wireframe with normals */
	if(SELECT == 1 || SELECT == 2)
	{
		SetLightsOff();
		glDisable(GL_TEXTURE_2D);
	}
    
	/* lighting on + smooth shading */
	if(SELECT == 3)
	{
		SetSmoothShading();
		ActivateLight();
		SetupLights();
		SetLightsOn();
	}

	/* lighting on + flat shading */
	if(SELECT == 4)
	{
		SetFlatShading();
		ActivateLight();
		SetupLights();
		SetLightsOn();
	}

	/* lighting off */
	if(SELECT == 5)
	{
		SetLightsOff();
	}

	/* textures on */
	if(SELECT == 6)
	{
		SetModeTextured();
	}

	/* textures off */
	if(SELECT == 7)
	{
		glDisable(GL_TEXTURE_2D);
	}

	/* switch to full screen mode */
	if(SELECT == 8)
	{
		glutFullScreen();
	}

	/* activate the wind for the cloths */
	if(SELECT == 9)
	{
		clothInstance->wind = 2.9;
		clothInstance->clothPhysics();
		clothInstance->initUnitForce();
	}

	/* stop the wind blowing on the cloths */
	if(SELECT == 10)
	{
		clothInstance->wind = 0.0;
		clothInstance->clothPhysics();
		clothInstance->initUnitForce();
	}
}

/*
=================
ControlMenu
-Used for the right-click accessible mouse Menu
=================
*/
void ControlMenu(int x)
{
	/* basic switch control structure, sets the global SELECT var responsible for
	   switching between the different rendering modes */
	switch(x)
	{
		case 1: SELECT = 1;
			break;
		case 2: SELECT = 2;
			break;
		case 3: SELECT = 3;
			break;
		case 4: SELECT = 4;
			break;
		case 5: SELECT = 5;
			break;
		case 6: SELECT = 6;
			break;
		case 7: SELECT = 7;
			break;
		case 8: SELECT = 8;
			break;
		case 9: SELECT = 9;
			break;
		case 10: SELECT = 10;
			break;
		case 11: SELECT = 11;
			break;
		case 12: SELECT = 12;
			break;
		case 13: exit(0);
	}
	/* sets the current window for redisplay */
	glutPostRedisplay();
}

/*
=================
Menu
-The mouse Menu GLUT routines
=================
*/
void Menu()
{
	/* creates a new pop-up Menu */
	glutCreateMenu(ControlMenu);
	/* adds Menu entries to the bottom of the current Menu */
	glutAddMenuEntry("Draw Wireframe",1);
	glutAddMenuEntry("Draw Wireframe with Normals",2);
	glutAddMenuEntry("Lighting On (Smooth Shading)",3);
	glutAddMenuEntry("Lighting On (Flat Shading)",4);
	glutAddMenuEntry("Lighting Off",5);
	glutAddMenuEntry("Textures On",6);
	glutAddMenuEntry("Textures Off",7);
	glutAddMenuEntry("Fullscreen",8);
	glutAddMenuEntry("Blow Wind",9);
	glutAddMenuEntry("Stop Wind",10);
	glutAddMenuEntry("Pool Transparency Off", 11);
	glutAddMenuEntry("Pool Transparency On", 12);
	glutAddMenuEntry("Quit",13);
	/* attaches the right mouse button for Menu selection */
	glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*
=================
main
=================
*/
int main(int argc, char** argv)
{
	/* initialise the GLUT library. */
	glutInit(&argc, argv);
	/* sets initial Display mode: double buffering, depth buffer, alpha masking */
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ALPHA);
	/* sets default window size */
	glutInitWindowSize(1024, 768);
	/* creates a top-level window with a caption */
	glutCreateWindow("An Interactive OpenGL Environment");
	/* Reshape callback for the current window -> for use in Reshape() */
	glutReshapeFunc(Reshape);
	/* displays our contents in the window */
	glutDisplayFunc(Display);
	/* sets special Keys callback for movement control Keys */
	glutSpecialFunc(InputKeyPressed);
	/* Keys handling */
	glutKeyboardFunc(Keys);
	/* sets the global Idle callback for background processing/animation */
	glutIdleFunc(Idle);
	//glewInit(); //only used in Linux/Unix implementation

	/* setups the cloth */
	clothInstance->particleSetup();
	clothInstance->clothPhysics();
	
	/* calls our Menu function */
	Menu();
	/* calls MyInit function */
	MyInit();

	/* enters the GLUT event processing loop - will handle any callback */
	glutMainLoop();
	return 0;
}
//EOF